/*
    Copyright (C) 2004  Erik Hjortsberg

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "EmberEntity.h"

#include "framework/Service.h"
#include "framework/ConsoleBackend.h"
#include "services/EmberServices.h"
#include "services/sound/SoundService.h"
#include "EmberEntityFactory.h"
#include "MotionManager.h"
#include "GUIManager.h"
#include "terrain/TerrainArea.h"
#include "terrain/TerrainMod.h"
#include "MathConverter.h"

#include "EmberOgre.h"
#include <OgreWireBoundingBox.h>
#include <OgreException.h>

#include "terrain/TerrainGenerator.h"
#include "terrain/TerrainPage.h"

#include <Mercator/Area.h>
#include <Mercator/TerrainMod.h>
//#include <Atlas/Objects/ObjectsFwd.h>
#include <Eris/TypeInfo.h>
#include <Eris/View.h>
#include <Atlas/Formatter.h>
#include <Atlas/Objects/Decoder.h>
#include <Atlas/Codecs/XML.h>
#include <Atlas/Message/MEncoder.h>
#include <Atlas/Message/QueuedDecoder.h>


#include "model/Model.h"

using namespace Ogre;


namespace Ogre {

	/**
	This is just like a WireBoundBox but not aligned to the axes, hence it will correctly line up according to it's orientation.
	*/
	class OOBBWireBoundingBox  : public WireBoundingBox
	{
		public:
	
		void getWorldTransforms( Matrix4* xform ) const
		{
			SimpleRenderable::getWorldTransforms(xform);
		}
	};

};














namespace EmberOgre {


const std::string EmberEntity::MODE_STANDING("standing");
const std::string EmberEntity::MODE_RUNNING("running");
const std::string EmberEntity::MODE_WALKING("walking");
const std::string EmberEntity::MODE_SWIMMING("swimming");
const std::string EmberEntity::MODE_FLOATING("floating");
const std::string EmberEntity::MODE_FIXED("fixed");

const std::string EmberEntity::BboxMaterialName("BaseYellowNoLightning");


EmberEntity::EmberEntity(const std::string& id, Eris::TypeInfo* ty, Eris::View* vw,Ogre::SceneManager* sceneManager)
:
Eris::Entity(id, ty, vw) 
, mIsInitialized(false)
, mIsInMotionManager(false)
, mErisEntityBoundingBox(0)
, mOgreNode(0)
, mTerrainArea(0)
, mTerrainMod(0)
, mMovementMode(MM_DEFAULT)
{
	createSceneNode(sceneManager);
}

EmberEntity::~EmberEntity()
{
	///detach all children so we don't destroy them
	while (getSceneNode()->numChildren()) {
		getSceneNode()->removeChild((short unsigned int)0);
	}
	Ogre::SceneNode* parent = static_cast<Ogre::SceneNode*>(getSceneNode()->getParent());
	if (parent) {
		parent->removeAndDestroyChild(getSceneNode()->getName());
	} else {
		getSceneNode()->getCreator()->destroySceneNode(getSceneNode()->getName());
	}
	///make sure it's not in the MotionManager
	///TODO: keep a marker in the entity so we don't need to call this for all entities
	MotionManager::getSingleton().removeEntity(this);

	if (mErisEntityBoundingBox) {
		mErisEntityBoundingBox->getParentSceneNode()->getCreator()->destroySceneNode(mErisEntityBoundingBox->getParentSceneNode()->getName());
	}
	OGRE_DELETE mErisEntityBoundingBox;
	
	//mSceneManager->destroySceneNode(getSceneNode()->getName());
}

void EmberEntity::init(const Atlas::Objects::Entity::RootEntity &ge, bool fromCreateOp)
{
	Eris::Entity::init(ge, fromCreateOp);
	
	synchronizeWithServer();
	
	// set the Ogre node position and orientation based on Atlas data
	std::stringstream ss;
	if (getPredictedPos().isValid()) {
		ss << "Entity " << getId() << "(" << getName() << ") placed at (" << getPredictedPos().x() << "," << getPredictedPos().y() << "," << getPredictedPos().x() << ")";
	}
	S_LOG_VERBOSE( ss.str());
	
	mIsInitialized = true;
	
	///Delay the checking and creation of area and terrainmod, since if we do it in onAttrChanged before the entity is properly initialized we'll get problems with multiple reparsings, as well as problems with placing the areas or terrainmod before the entity has been moved to it's proper place.
	///Another way of doing this would be by attaching listeners, but that require more memory. It might be in the end that that's a better approach, depending on how much memory is needed, contrasted with the extra cycles spent here for every entity.
	if (hasAttr("area")) {
		createDependentObject("area");
	}
	
	if (hasAttr("terrainmod")) {
		createDependentObject("terrainmod");
	}

}

bool EmberEntity::createDependentObject(const std::string& attributeName)
{
	///if the area attribute has changed, and we _don't_ have any mTerrainArea instance, try to create one such.
	///if we do have an mTerrainArea instance, all updates will be taken care of by that instead and we can ignore this
	if (attributeName == "area" && !mTerrainArea.get()) {
		mTerrainArea = std::auto_ptr<Terrain::TerrainArea>(new Terrain::TerrainArea(this));
		if (mTerrainArea->init()) {
			addArea(mTerrainArea.get());
			return true;
		} else {
			///if we couldn't properly initialize, delete the instance now, and then hopefully the next time the "area" attribute is changed we'll be able to properly create an area
			mTerrainArea.reset();
		}
	}
	
	///if the area attribute has changed, and we _don't_ have any mTerrainMod instance, try to create one such.
	///if we do have an mTerrainMod instance, all updates will be taken care of by that instead and we can ignore this
	if (attributeName == "terrainmod" && !mTerrainMod.get()) {
		mTerrainMod = std::auto_ptr<Terrain::TerrainMod>(new Terrain::TerrainMod(this));
		if (mTerrainMod->init()) {
			addTerrainMod(mTerrainMod.get());
			return true;
		} else {
			///if we couldn't properly initialize, delete the instance now, and then hopefully the next time the "area" attribute is changed we'll be able to properly create a mod
			mTerrainMod.reset();
		}
	}
	return false;
}


void EmberEntity::synchronizeWithServer()
{
	if (getPosition().isValid()) {
		getSceneNode()->setPosition(Atlas2Ogre(getPredictedPos()));
		adjustPosition();
	}
	if (getOrientation().isValid()) {
		getSceneNode()->setOrientation(Atlas2Ogre(getOrientation()));
	}
}

void EmberEntity::createSceneNode(Ogre::SceneManager* sceneManager)
{
	EmberEntity* container = getEmberLocation();
	if (container == 0) {
		mOgreNode = sceneManager->createSceneNode(getId());
	} else {
		Ogre::SceneNode * node = container->getSceneNode();
		mOgreNode = node->createChildSceneNode(getId());
	}
}

void EmberEntity::updateMotion(Ogre::Real timeSlice)
{
	getSceneNode()->setPosition(Atlas2Ogre(getPredictedPos()));
	adjustPosition();
	
	//if there's a debug bounding box for the eris entity, update it's position
	if (mErisEntityBoundingBox) {
		mErisEntityBoundingBox->getParentSceneNode()->setPosition(getSceneNode()->getPosition());
		mErisEntityBoundingBox->getParentSceneNode()->setOrientation(getSceneNode()->getOrientation());
	}

}

void EmberEntity::onMoved()
{
	Eris::Entity::onMoved();
	const WFMath::Quaternion& orient = getOrientation();
	getSceneNode()->setOrientation(Atlas2Ogre(orient));
	updateMotion(0);
}

void EmberEntity::setMoving(bool moving)
{
	// Call the overridden method 
	Eris::Entity::setMoving(moving);
	
	MotionManager* motionManager = &MotionManager::getSingleton();
	if (moving) {
		//the entity is moving
		if (!mIsInMotionManager) {
			motionManager->addEntity(this);
			mIsInMotionManager = true;
		}
	} else {
		//the entity has stopped moving
		if (mIsInMotionManager) {
			motionManager->removeEntity(this);
			mIsInMotionManager = false;
		}
	}
}

void EmberEntity::onTalk(const Atlas::Objects::Operation::RootOperation& talkArgs)
{
	const std::vector<Atlas::Objects::Root>& args = talkArgs->getArgs();
	if (args.empty()) {
        Eris::Entity::onTalk(talkArgs);
		return;
	}

	const Atlas::Objects::Root& talk = args.front();
	
	
    if (!talk->hasAttr("say")) {
        Eris::Entity::onTalk(talkArgs);
		return;
    }


	///some talk operations come with a predefined set of suitable responses, so we'll store those so that they can later on be queried by the GUI for example
	mSuggestedResponses.clear();
	if (talk->hasAttr("responses")) {
		if (talk->getAttr("responses").isList()) {
			const Atlas::Message::ListType & responseList = talk->getAttr("responses").asList();
			Atlas::Message::ListType::const_iterator I = responseList.begin();
			for(; I != responseList.end(); ++I) {
				mSuggestedResponses.push_back(I->asString());
			}
		
		}
	}
	
	const std::string& msg = talk->getAttr("say").asString();
    


	
	std::string message = "<";
	message.append(getName());
	message.append(",");
	const std::string& type = getType()->getName(); // Eris type as a string
	message.append(type);
	message.append("> ");
	message.append(msg);
	S_LOG_VERBOSE( "Entity says: [" << message << "]\n" );

	/// Make the message appear in the chat box
	GUIManager::getSingleton().AppendIGChatLine.emit(msg, this);

	/// Make a sound in OpenAL
//	Ember::EmberServices::getSingleton().getSoundService()->playTalk(msg,
//		getPosition(),getOrientation());

	/// Call the method of the base class (since we've overloaded it)
	Eris::Entity::onTalk(talkArgs);
}


void EmberEntity::onSoundAction( const Atlas::Objects::Operation::RootOperation & op )
{
	///We'll just catch the call and write something to both the log and the console, and then pass it on.
	const std::list<std::string> &p = op->getParents();
	std::list<std::string>::const_iterator I = p.begin();
	
	if (I != p.end()) {
	
		const std::string& name = *I;
		std::string message = getName() + " emits a " + name + ".";
		
		Ember::ConsoleBackend::getSingletonPtr()->pushMessage(message);
		
		S_LOG_VERBOSE( "Entity: " << this->getId() << " (" << this->getName() << ") sound action: " << name);
	}
	
	Eris::Entity::onSoundAction(op);
}


void EmberEntity::onVisibilityChanged(bool vis)
{
	checkClientVisibility(vis);
	Eris::Entity::onVisibilityChanged(vis);
}

void EmberEntity::checkClientVisibility(bool vis)
{
	///since we don't want to show all entities solely by their server flags (for example, an inventory item belonging to a character might not be shown even though the server thinks it's visible) we have to some more checks before we decide whether to show this or not
	EmberEntity* container = static_cast<EmberEntity*>(getLocation());
	if (container) {
		///check with the parent first if we should show ourselves
		if (vis && container->allowVisibilityOfMember(this)) {
			///don't cascade, only change the top node
			setClientVisible(true);	
		} else {
			setClientVisible(false);	
		}
		
	} else {
		setClientVisible(vis);
	}
}


void EmberEntity::setClientVisible(bool visible)
{
	///when entities are hidden, we detach them from the rendering scene graph altogether. this speeds up Ogre since it doesn't have to calculate visibility for nodes that are hidden anyway
	if (!visible) {
		if (getSceneNode()->getParent()) {
			getSceneNode()->getParent()->removeChild(getSceneNode());
		}
	} else {
		if (getLocation()) {
			if (!getSceneNode()->getParent()) {
				getEmberLocation()->getSceneNode()->addChild(getSceneNode());
			}
		}
	}
	EventClientVisibilityChanged.emit(visible);
// 	getSceneNode()->setVisible(visible && getLocation(), false);	
}


void EmberEntity::adjustPosition()
{
	if (getPredictedPos().isValid()) {
		adjustPosition(Atlas2Ogre( getPredictedPos() ));
	}
}

void EmberEntity::adjustPosition(const Ogre::Vector3& position)
{
	if (mMovementMode == MM_FIXED) {

	} else {
		EmberEntity* container = getEmberLocation();
		if (container) {
			container->adjustPositionForContainedNode(this, position);
		}
	}	
}

const Ogre::Vector3& EmberEntity::getOffsetForContainedNode(const Ogre::Vector3& localPosition, EmberEntity* const entity)
{
	///send it upwards until we get a an entity which knows how to set the position (we'll in most cases end up in the world which will adjust the height a bit)
	EmberEntity* container = getEmberLocation();
	if (container) {
		//TerrainPosition derivedPosition(getPredictedPos().x() + position.x(), getPredictedPos().y() + position.y());
		return container->getOffsetForContainedNode(localPosition + getSceneNode()->getPosition(), entity);
	} else {
		return Ogre::Vector3::ZERO;
	}
	
	
}



void EmberEntity::adjustPositionForContainedNode(EmberEntity* const entity, const Ogre::Vector3& position) 
{

	Ogre::SceneNode* sceneNode = entity->getSceneNode();
	//Ogre::Vector3 position = sceneNode->getPosition();
	const Ogre::Vector3& offset = getOffsetForContainedNode(position, entity);
	if (offset != Ogre::Vector3::ZERO) {
        
		sceneNode->setPosition(position + offset);
	}
	
}

void EmberEntity::onLocationChanged(Eris::Entity *oldLocation)
{
	
	if (getLocation() == oldLocation)
	{
		///If it's the same location, don't do anything, but do add a warning to the log since this isn't supposed to happen.
		S_LOG_WARNING( "Same new location as old for entity: " << this->getId() << " (" << this->getName() << ")" );
		return Eris::Entity::onLocationChanged(oldLocation);
	}
	Eris::Entity::onLocationChanged(oldLocation);
	
	///Ff we're attached to something, detach from it.
	detachFromModel();

	///Before we detach ourselves from our current parent, we need to record our current position in the world. This will come in handy later on when we need to determine if we actually moved in the world space.
	const Ogre::Vector3 oldWorldPosition = getSceneNode()->_getDerivedPosition();
	
	///Get the new location. We use getEmberLocation() since we always know that all entities are of type EmberEntity.
	EmberEntity* newLocationEntity = getEmberLocation();
	if (getSceneNode()->getParentSceneNode()) {
		///Detach from our current parent scene node, removing us from the scene node graph.
		getSceneNode()->getParentSceneNode()->removeChild(getSceneNode()->getName());
	}
	if (!newLocationEntity) {
		///If there's no new location, just leave now. We've already removed ourselves from the scene node graph, so this entity is in effect in limbo.
		return;	
	} else {
		
		///Add ourselves to the scene node of our parent.
		newLocationEntity->getSceneNode()->addChild(getSceneNode());
		S_LOG_VERBOSE( "Entity: " << this->getId() << " (" << this->getName() << ") relocated to: "<< newLocationEntity->getId() << " (" << newLocationEntity->getName() << ")" );
		if (getPosition().isValid()) {
			///Note that in some instances, for instance when the avatar enters the sty, the position isn't updated yet, which will make the avatar "snap" to an incorrect position (since the parent node has changed) until next frame, when the position should have been updated.
			getSceneNode()->setPosition(Atlas2Ogre(getPredictedPos()));
			adjustPosition();
			std::stringstream ss;
			ss << getPredictedPos();
			S_LOG_VERBOSE("New position for entity: "  << this->getId() << " (" << this->getName() << " ) :" << ss.str());
		}
		if (getOrientation().isValid()) {
			getSceneNode()->setOrientation(Atlas2Ogre(getOrientation()));
			std::stringstream ss;
			ss << getOrientation();
			S_LOG_VERBOSE("New orientation for entity: "  << this->getId() << " (" << this->getName() << " ) :" << ss.str());
		}
	
		checkClientVisibility(isVisible());
	
		///we'll adjust the entity so it retains it's former position in the world, but only for moving entities
		///since else we'll get a "gap" when we're waiting on updated positions from the server
		///this isn't optimal
		if (isMoving()) {
			const Ogre::Vector3& newWorldPosition = getSceneNode()->_getDerivedPosition();
			getSceneNode()->translate(oldWorldPosition - newWorldPosition);
		}
	}
	
}

void EmberEntity::onAction(const Atlas::Objects::Operation::RootOperation& act)
{
	const std::list<std::string> &p = act->getParents();
	std::list<std::string>::const_iterator I = p.begin();
	
	if (I != p.end()) {
	
		const std::string& name = *I;
		std::string message = getName() + " performs a " + name + ".";
		
		Ember::ConsoleBackend::getSingletonPtr()->pushMessage(message);
		
		S_LOG_VERBOSE( "Entity: " << this->getId() << " (" << this->getName() << ") action: " << name);
	}
	Entity::onAction(act);
}

void EmberEntity::onImaginary(const Atlas::Objects::Root& act)
{
    Atlas::Message::Element attr;
    if (act->copyAttr("description", attr) && attr.isString()) {
		std::string message = getName() + " " + attr.asString() + ".";
		
		Ember::ConsoleBackend::getSingletonPtr()->pushMessage(message);
			
		S_LOG_VERBOSE("Entity: " << this->getId() << " (" << this->getName() << ") imaginary: " << attr.String());
    }

	Entity::onImaginary(act);

}

bool EmberEntity::allowVisibilityOfMember(EmberEntity* entity) {
	return true;
}

const std::vector< std::string >& EmberEntity::getSuggestedResponses( ) const
{
	return mSuggestedResponses;
}

bool EmberEntity::hasSuggestedResponses( ) const
{
	return mSuggestedResponses.size() > 0;
}


void EmberEntity::addArea(Terrain::TerrainArea* area)
{
	///A normal EmberEntity shouldn't know anything about the terrain, so we can't handle the area here.
	///Intead we just pass it on to the parent until we get to someone who knows how to handle this (preferrably the terrain).
	if (getEmberLocation()) {
		getEmberLocation()->addArea(area);
	}
}

void EmberEntity::addTerrainMod(Terrain::TerrainMod* mod)
{
///Same as addArea: pass it on to the parent until it gets to someone who knows how to handle this
    if (getEmberLocation()) {
        getEmberLocation()->addTerrainMod(mod);
    }
}

void EmberEntity::onAttrChanged(const std::string& str, const Atlas::Message::Element& v)
{
    if (str == "mode") {
    	parseModeChange(v);
	} else if (str == "bbox") {
		Entity::onAttrChanged(str, v);
		onBboxChanged();
		return;
	}
	
	///call this before checking for areas and terrainmods, since those instances created to handle that (TerrainMod and TerrainArea) will listen to the AttrChanged event, which would then be emitted after those have been created, causing duplicate regeneration from the same data
	Entity::onAttrChanged(str, v);
	
	///Only to this if the entity has been propely initialized, to avoid using incomplete entity data (for example an entity which haven't yet been moved to it's proper place, as well as avoiding duplicate parsing of the same data.
	if (mIsInitialized) {
		createDependentObject(str);
	}

	
}

void EmberEntity::parseModeChange(const Atlas::Message::Element& v)
{
	const std::string& mode = v.asString();
	MovementMode newMode;
	if (mode.empty()) {
		newMode = MM_DEFAULT;
	} else if (mode == MODE_STANDING) {
		newMode = MM_STANDING;
	} else if (mode == MODE_RUNNING) {
		newMode = MM_RUNNING;
	} else if (mode == MODE_WALKING) {
		newMode = MM_WALKING;
	} else if (mode == MODE_SWIMMING) {
		newMode = MM_SWIMMING;
	} else if (mode == MODE_FLOATING) {
		newMode = MM_FLOATING;
	} else if (mode == MODE_FIXED) {
		newMode = MM_FIXED;
	} else {
		newMode = MM_DEFAULT;
	}
	
	onModeChanged(newMode);
}

void EmberEntity::onModeChanged(MovementMode newMode)
{
	if (newMode == MM_FIXED) {
		adjustPosition();
	}
	EventModeChanged.emit(newMode);
	mMovementMode = newMode;
}

void EmberEntity::setVisualize(const std::string& visualization, bool visualize)
{
	if (visualization == "OgreBBox") {
		showOgreBoundingBox(visualize);
	} else if (visualization == "ErisBBox") {
		showErisBoundingBox(visualize);
	}
}

bool EmberEntity::getVisualize(const std::string& visualization) const
{
	if (visualization == "OgreBBox") {
		return getShowOgreBoundingBox();
	} else if (visualization == "ErisBBox") {
		return getShowErisBoundingBox();
	}
	return false;
}


void EmberEntity::showOgreBoundingBox(bool show)
{
	getSceneNode()->showBoundingBox(show);
}

void EmberEntity::showErisBoundingBox(bool show)
{

	createErisBboxMaterial();
	
	///if there's no bounding box, create one now
	///allowing for some lazy loading
	if (!mErisEntityBoundingBox) {
		mErisEntityBoundingBox = OGRE_NEW Ogre::OOBBWireBoundingBox();
		mErisEntityBoundingBox->setMaterial(BboxMaterialName);
		Ogre::SceneNode* boundingBoxNode = EmberOgre::getSingleton().getWorldSceneNode()->createChildSceneNode();
		boundingBoxNode->attachObject(mErisEntityBoundingBox);
		
		///if there's no bounding box defined for this entity, show one that is 0.2 meters across in all direction
		if (hasBBox()) {
			mErisEntityBoundingBox->setupBoundingBox(Atlas2Ogre(getBBox()));
		} else {
			mErisEntityBoundingBox->setupBoundingBox(Ogre::AxisAlignedBox(-0.1, -0.1, -0.1, 0.1, 0.1, 0.1));
		}
		
		boundingBoxNode->setPosition(Atlas2Ogre(getPredictedPos()));
		boundingBoxNode->setOrientation(Atlas2Ogre(getOrientation()));
	}
	mErisEntityBoundingBox->setVisible(show);
	
}

void EmberEntity::createErisBboxMaterial()
{
	if (!Ogre::MaterialManager::getSingleton().resourceExists(BboxMaterialName)) {
		Ogre::MaterialPtr baseYellowNoLighting = Ogre::MaterialManager::getSingleton().create(BboxMaterialName, 
			Ogre::ResourceGroupManager::INTERNAL_RESOURCE_GROUP_NAME);
 		baseYellowNoLighting->setLightingEnabled(false);
		baseYellowNoLighting->setAmbient(Ogre::ColourValue(1, 1, 0.7));
		baseYellowNoLighting->setDiffuse(Ogre::ColourValue(1, 1, 0.7));
	}
}

void EmberEntity::onBboxChanged()
{
	if (mErisEntityBoundingBox) {
		mErisEntityBoundingBox->setupBoundingBox(Atlas2Ogre(getBBox()));
	}
}


bool EmberEntity::getShowOgreBoundingBox() const
{
	return getSceneNode()->getShowBoundingBox();
}
bool EmberEntity::getShowErisBoundingBox() const
{
	return (mErisEntityBoundingBox && mErisEntityBoundingBox->isVisible());
	
}

Ogre::SceneNode* EmberEntity::getSceneNode() const 
{
	return mOgreNode;	
}

EmberEntity* EmberEntity::getEmberLocation() const
{ 
	return static_cast<EmberEntity*>(getLocation());
}


const Ogre::AxisAlignedBox& EmberEntity::getWorldBoundingBox(bool derive) const
{
	static Ogre::AxisAlignedBox boundingBox(0,0,0,0,0,0);
	return boundingBox;
}
	
const Ogre::Sphere & EmberEntity::getWorldBoundingSphere (bool derive) const
{
	static Ogre::Sphere sphere;
	return sphere;
}

std::vector<std::string> EmberEntity::getActions()
{
	///get the actions from Eris and return them a simple vector of strings
	std::vector<std::string> actions;
	
	if (hasAttr("actions")) {
		const Atlas::Message::Element& operations = valueOfAttr("actions");
		if (operations.isList()) {
			const Atlas::Message::ListType& list = operations.asList();
			actions.reserve(list.size());
			Atlas::Message::ListType::const_iterator J = list.begin();
			for (; J != list.end(); ++J) {
				if (J->isString()) {
					actions.push_back(J->asString());
				}
			}
		}
	}

	return actions;
}

std::vector<std::string> EmberEntity::getDefaultUseOperators()
{
	///get the use operations from Eris and return them a simple vector of strings
	std::vector<std::string> operators;
	
	Eris::TypeInfoArray types = getUseOperations();
	
	for (Eris::TypeInfoArray::iterator I = types.begin(); I != types.end(); ++I) {
		operators.push_back((*I)->getName());
	}
	
	return operators;
}

Ogre::SceneManager* EmberEntity::getSceneManager()
{
	assert(mOgreNode);
	return mOgreNode->getCreator();
}

void EmberEntity::dumpAttributes(std::iostream& outstream, std::ostream& logOutstream) const 
{
	logOutstream << "Dumping attributes for entity " << getId() << "(" << getName() << ")" << std::endl;
	
	Atlas::Message::QueuedDecoder decoder;
	//std::fstream file;
	
	Atlas::Codecs::XML codec(outstream, decoder);
	Atlas::Formatter formatter(outstream, codec);
	Atlas::Message::Encoder encoder(formatter);
	formatter.streamBegin();
	encoder.streamMessageElement(getAttributes());
		
	formatter.streamEnd();
}

}





